# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

import tobiko
from tobiko.tests import unit


LOG = log.getLogger(__name__)


class CapureLogTest(unit.TobikoUnitTest):

    def test_capture_log_content(self):
        with tobiko.CaptureLogFixture(test_case_id=self.id()) as capture:
            LOG.warning('< Some debug line>')
        logged_lines = capture.getDetails()['log'].as_text().splitlines()
        expected = f" WARNING {__name__} | < Some debug line>"
        for line in logged_lines:
            if expected in line:
                break
        else:
            lines = '\n'.join(logged_lines)
            self.fail(f"log line not captured: '{expected}' not in \n"
                      f"{lines}")
